'use strict';
const http2 = require('http2');
const {Writable} = require('stream');
const HTTP2IncomingMessage = require('./HTTP2IncomingMessage');
const {
	ERR_HTTP2_HEADERS_SENT,
	ERR_HTTP2_PSEUDOHEADER_NOT_ALLOWED,
	ERR_INVALID_HTTP_TOKEN,
	ERR_HTTP2_INVALID_HEADER_VALUE,
	createHangUpError
} = require('./errors');

const {
	HTTP2_HEADER_AUTHORITY,
	HTTP2_HEADER_METHOD,
	HTTP2_HEADER_PATH,
	HTTP2_HEADER_SCHEME,
	HTTP2_HEADER_STATUS,
	NGHTTP2_CANCEL
} = http2.constants;

const isPseudoHeader = name => {
	switch (name) {
		case HTTP2_HEADER_STATUS:		// :status
		case HTTP2_HEADER_METHOD:		// :method
		case HTTP2_HEADER_PATH:			// :path
		case HTTP2_HEADER_AUTHORITY:	// :authority
		case HTTP2_HEADER_SCHEME:		// :scheme
			return true;
		default:
			return false;
	}
};

/* istanbul ignore next */
const deferToConnect = (socket, method, ...args) => {
	let call;
	if (typeof method === 'function') {
		call = method;
	} else {
		call = () => socket[method](...args);
	}

	if (socket.writable) {
		call();
	} else {
		socket.once('connect', call);
	}
};

const kBody = Symbol('body');
const kHeaders = Symbol('headers');
const kKeepAlive = Symbol('keepAlive');

class HTTP2ClientRequest extends Writable {
	constructor(options, cb) {
		super();

		this[kBody] = [];
		this[kHeaders] = {};

		const self = this;
		const port = options.port || (options.protocol === 'https:' ? 443 : 80);

		if (options.session) {
			this.session = options.session;
		} else {
			let authority;
			if (options.headers && options.headers.authority) {
				authority = options.headers.authority; // eslint-disable-line prefer-destructuring
				delete options.headers.authority;
			} else {
				authority = `${options.protocol}//${options.hostname || options.host}:${port}`;
			}

			this.session = http2.connect(authority);
		}

		// Haha, this function isn't proxied, so we can get raw socket here! Workaround for #2
		this.socket = this.session.socket.ref();
		this.connection = this.socket;

		this.session.on('error', err => self.emit('error', err));
		this.socket.on('close', () => {
			if (self.res) {
				if (!self.res.complete) {
					self.res.aborted = true;
					self.res.emit('aborted');
				}
				self.emit('close');
				if (self.res.readable) {
					self.res.on('end', () => {
						self.res.emit('close');
					});
					self.res.push(null);
				} else {
					self.res.emit('close');
				}
			} else {
				if (!self.socket._hadError) {
					self.socket._hadError = true;
					self.emit('error', createHangUpError());
				}
				self.emit('close');
			}
		});

		this.method = options.method || 'GET';
		this.path = options.path || '/';
		if (cb) {
			this.once('response', cb);
		}

		this.once('finish', () => {
			if (this.session.destroyed) {
				return;
			}

			if (!this._request) {
				this.flushHeaders();
			}
			this._request.end();
		});

		this.res = null;
		this.aborted = false;
		this[kKeepAlive] = false;

		if (options.headers) {
			for (const [header, value] of Object.entries(options.headers)) {
				this.setHeader(header, value);
			}
		}

		process.nextTick(() => self.emit('socket', this.socket));
	}

	_write(chunk, enc, next) {
		if (this._request) {
			this._request.write(chunk, enc);
		} else {
			this[kBody].push(Buffer.from(chunk, enc));
		}
		next();
	}

	abort() {
		const self = this;
		if (!this.aborted) {
			process.nextTick(() => self.emit('abort'));
		}
		this.aborted = true;

		if (this.res) {
			this.res._dump();
		}

		if (this._request) {
			this._request.close(NGHTTP2_CANCEL);
		} else {
			this.session.close();
		}
	}

	flushHeaders() {
		if (!this._request) {
			const self = this;
			this._request = this.session.request({
				[HTTP2_HEADER_PATH]: this.path,
				[HTTP2_HEADER_METHOD]: this.method,
				...this[kHeaders]
			});

			const proxyEvent = e => this._request.on(e, (...args) => {
				self.emit(e, ...args);
			});

			[
				'timeout',
				'continue',
				'error'
			].forEach(proxyEvent);

			this._request.on('response', (headers, flags, rawHeaders) => {
				this.res = new HTTP2IncomingMessage(self, {headers, flags, rawHeaders});
				if (!self.emit('response', this.res)) {
					this.res._dump();
				}
			});

			this._request.on('trailers', (trailers, flags, rawTrailers) => {
				self.res.trailers = trailers;
				self.res.rawTrailers = rawTrailers;
			});

			this._request.once('close', () => self.session.close());

			for (const element of this[kBody]) {
				this._request.write(element);
			}

			this.emit('flushedHeaders');
		}
	}

	getHeader(name) {
		return this[kHeaders][name];
	}

	get headersSent() {
		if (this._request) {
			return Boolean(this._request.sentHeaders);
		}

		return false;
	}

	removeHeader(name) {
		if (this.headersSent) {
			throw new ERR_HTTP2_HEADERS_SENT('remove');
		}

		delete this[kHeaders][name];
	}

	setHeader(name, value) {
		if (this.headersSent) {
			throw new ERR_HTTP2_HEADERS_SENT('set');
		}
		if (name === '' || typeof name !== 'string') {
			throw new ERR_INVALID_HTTP_TOKEN('Header name', name);
		}
		name = name.trim().toLowerCase();
		if (isPseudoHeader(name)) {
			throw new ERR_HTTP2_PSEUDOHEADER_NOT_ALLOWED();
		}
		if (value === undefined || value === null) {
			throw new ERR_HTTP2_INVALID_HEADER_VALUE(value, name);
		}

		this[kHeaders][name] = value;
	}

	setNoDelay(noDelay) {
		deferToConnect(this.socket, 'setNoDelay', noDelay);
	}

	setTimeout(ms, cb) {
		if (this._request) {
			this._request.setTimeout(ms, cb);
		} else {
			const self = this;
			this.once('flushedHeaders', () => {
				self._request.setTimeout(ms, cb);
			});
		}
		return this;
	}
}

module.exports = HTTP2ClientRequest;
